-----------------------------------------------------------------------------------------------
-- Client Lua Script for AutoScan
-- Copyright (c) Dennis Schmidt
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- AutoScan Module Definition
-----------------------------------------------------------------------------------------------
local AutoScan = {} 
local userSettings = {}
local defaultSettings = {
	autoScan = true,
	scanDistance = 10,
	scanOnTarget = true,
	autoSummon = true
}

-- copy defaults
for k,v in pairs(defaultSettings) do 
	userSettings[k] = v
end	
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function AutoScan:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    self.units = {}
	self.unknownUnits = {}
	
    self.playerUnit = nil
    self.scanStart = nil

	self.timer = nil
	self.unknownTimer = nil
	
    return o
end

function AutoScan:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- AutoScan OnLoad
-----------------------------------------------------------------------------------------------
function AutoScan:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("AutoScan.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("TargetUnitChanged","OnTargetUnitChanged", self)
end

-----------------------------------------------------------------------------------------------
-- AutoScan OnDocLoaded
-----------------------------------------------------------------------------------------------
function AutoScan:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "AutoScanForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
	    self.wndMain:Show(false, true)
		Apollo.RegisterSlashCommand("asc", "OnSlashCommand", self)
		Apollo.RegisterSlashCommand("autoscan", "OnSlashCommand", self)
		Apollo.RegisterSlashCommand("ascdebug", "OnDebugCommand", self)
		self.timer = ApolloTimer.Create(0.5, true, "OnTimer", self)
		self.unknownTimer = ApolloTimer.Create(5.0, true, "OnUnknownUnitsTimer", self)
	end
end

-- Save User Settings
function AutoScan:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
	local tSave = {}
    for k,_ in pairs(defaultSettings) do 
		tSave[k] = userSettings[k] 
	end	
	return tSave
end

-- Restore Saved User Settings
function AutoScan:OnRestore(eType, t)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then return end
    for k,v in pairs(defaultSettings) do 
		if t[k] ~= nil then 
			userSettings[k] = t[k]
		else
			userSettings[k] = v
 		end
	end	
end

function AutoScan:IsUnitScannable(unit)
	local isScannable = false
	if unit ~= nil and unit:IsValid() then
		for k,v in pairs(unit:GetRewardInfo() or {}) do
			if v.strType=="Scientist" then
				isScannable = true
				break
			end
		end
	end
	return isScannable
end

function AutoScan:AddScannableUnit(unit)
	local unitId = unit:GetId()
	self.unknownUnits[unitId] = nil
	local tUnit = {}
	tUnit.scanned = false
	tUnit.unit = unit
	self.units[unitId] = tUnit
	return tUnit
end

function AutoScan:OnUnitCreated(unit)
	local unitId = unit:GetId()
	-- Player units are definitely not scannable, so don't waste the memory
	if (unit:GetType() == "Player") then
		return
	end
	
	-- apparently, when units are created, SOMETIMES the RewardInfo is not present yet, wtf?
	-- so to preserve performance, we only want the units in self.units to be confirmed scannable
	-- otherwise the processing of units is excessive and consumes a lot of CPU/frame
	-- if we can confirm now that the unit is scannable, we immediately
	-- queue it for regular scanning.  If not, every so often, we will check again by
	-- saving the unit in unknownUnits and see if the rewardinfo has been added
	if self:IsUnitScannable(unit) then
		self:AddScannableUnit(unit)
	else
		self.unknownUnits[unitId] = unitId
	end
end

function AutoScan:OnUnitDestroyed(unit)
	local unitId = unit:GetId()
	self.units[unitId] = nil
	self.unknownUnits[unitId] = nil
end

function AutoScan:OnTargetUnitChanged(unit)
	if userSettings.scanOnTarget and unit and PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Scientist then
		if self:IsUnitScannable(unit) then
			-- check to make sure this unit has made its way into the self.units
			local unitId = unit:GetId()
			local tUnit = self.units[unitId]
			if tUnit == nil then
				tUnit = self:AddScannableUnit(unit)
			end			
			self:ScanUnit(tUnit)
		end
	end
end

function AutoScan:GetDistanceTo(uUnit)
	if self.playerUnit ~= nil then
		local tPos1 = self.playerUnit:GetPosition()
		local tPos2 = uUnit:GetPosition()
		if tPos1 ~= nil and tPos2 ~= nil then
			return math.sqrt(math.pow(tPos1.x - tPos2.x, 2) + math.pow(tPos1.y - tPos2.y, 2) + math.pow(tPos1.z - tPos2.z, 2))
		end
	end
	return 99999
end

function AutoScan:ScanUnit(tUnit)
	local t = GameLib.GetGameTime()
	if self.scanStart == nil or (t - self.scanStart) > 4 then
		self.scanStart = t
		GameLib.SetTargetUnit(tUnit.unit)
		PlayerPathLib.PathAction()
		tUnit.scanned = true
	end
end

-- checks if a unit has been identified as a scannable unit
function AutoScan:OnUnknownUnitsTimer()
	if PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Scientist then
		if self.playerUnit ~= nil and self.playerUnit:IsValid() then
			for unitId in pairs(self.unknownUnits) do
				local unit = GameLib.GetUnitById(unitId)
				if self:IsUnitScannable(unit) then
					self:AddScannableUnit(unit)
				end
			end
		end
	end
end

function AutoScan:OnTimer()
	if userSettings.autoScan and PlayerPathLib.GetPlayerPathType() == PlayerPathLib.PlayerPathType_Scientist then
		if self.playerUnit ~= nil and self.playerUnit:IsValid() then
			if (not PlayerPathLib.ScientistHasScanBot() and userSettings.autoSummon) then
				PlayerPathLib.ScientistToggleScanBot()
			end
			if PlayerPathLib.ScientistHasScanBot() then
				local stealth = false
				if self.playerUnit:GetClassId() == GameLib.CodeEnumClass.Stalker then
					local buffs = self.playerUnit:GetBuffs()
					if buffs and buffs.arBeneficial then
					    local spells = buffs.arBeneficial
						for i=1, #spells do
							local spell = spells[i]
							local spellId = spell.splEffect:GetId()
							local spellName = spell.splEffect:GetName()
							-- If this is a Stalker in Stealth
							if spellId == 38784 then
								spells = nil
								buffs = nil
								stealth = true
								break
							end
						end
					end
				end
				for unitId,tUnit in pairs(self.units) do
					if not tUnit.scanned then
						local unit = tUnit.unit
						local isHostile = unit:GetDispositionTo(self.playerUnit) == Unit.CodeEnumDisposition.Hostile
						if (stealth and not isHostile) or (not stealth) then
							local distance = self:GetDistanceTo(unit)
						    if  distance < userSettings.scanDistance then
								self:ScanUnit(tUnit)
								return -- the timer will catch the next one
							end
						end
					end
				end
			end
			
		else
			self.playerUnit = GameLib.GetPlayerUnit()
		end
	end
end
-----------------------------------------------------------------------------------------------
-- AutoScan Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- on SlashCommand "/asc"
function AutoScan:OnSlashCommand()
	self.wndMain:FindChild("AutomaticScan"):SetCheck(userSettings.autoScan)
	self.wndMain:FindChild("ScanOnTargetButton"):SetCheck(userSettings.scanOnTarget)
	self.wndMain:FindChild("AutoSummon"):SetCheck(userSettings.autoSummon)
	self.wndMain:FindChild("ScanDistanceEditBox"):SetText(userSettings.scanDistance)
	self.wndMain:Invoke() -- show the window
end

function AutoScan:OnDebugCommand()
	SendVarToRover("_autoscan.self.units",self.units)
	SendVarToRover("_autoscan.self.unknownUnits",self.unknownUnits)
	
end

-----------------------------------------------------------------------------------------------
-- AutoScanForm Functions
-----------------------------------------------------------------------------------------------
-- when the Cancel button is clicked
function AutoScan:OnCancel()
	self.wndMain:Close() -- hide the window
end

function AutoScan:OnAutomaticScan( wndHandler, wndControl, eMouseButton )
	userSettings.autoScan = wndControl:IsChecked()
end

function AutoScan:OnScanTarget( wndHandler, wndControl, eMouseButton )
	userSettings.scanOnTarget = wndControl:IsChecked()
end

function AutoScan:OnScanDistanceReturn( wndHandler, wndControl, strText )
	userSettings.scanDistance = tonumber(strText)
end

function AutoScan:OnAutoSummon( wndHandler, wndControl, eMouseButton )
	userSettings.autoSummon = wndControl:IsChecked()
end

-----------------------------------------------------------------------------------------------
-- AutoScan Instance
-----------------------------------------------------------------------------------------------
local AutoScanInst = AutoScan:new()
AutoScanInst:Init()
